@@ -17,25 +17,33 @@
 
 package org.apache.commons.csv;
 
+import static org.apache.commons.csv.Token.Type.TOKEN;
+
 import java.io.Closeable;
 import java.io.File;
-import java.io.FileInputStream;
 import java.io.IOException;
+import java.io.InputStream;
 import java.io.InputStreamReader;
 import java.io.Reader;
 import java.io.StringReader;
 import java.net.URL;
 import java.nio.charset.Charset;
+import java.nio.file.Files;
+import java.nio.file.Path;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Collections;
 import java.util.Iterator;
 import java.util.LinkedHashMap;
 import java.util.List;
 import java.util.Map;
 import java.util.NoSuchElementException;
+import java.util.Objects;
+import java.util.Spliterator;
+import java.util.Spliterators;
 import java.util.TreeMap;
-
-import static org.apache.commons.csv.Token.Type.*;
+import java.util.stream.Stream;
+import java.util.stream.StreamSupport;
 
 /**
  * Parses CSV files according to the specified format.
@@ -96,7 +104,7 @@ import static org.apache.commons.csv.Token.Type.*;
  *
  * <p>
  * If the predefined formats don't match the format at hands, custom formats can be defined. More information about
- * customising CSVFormats is available in {@link CSVFormat CSVFormat JavaDoc}.
+ * customising CSVFormats is available in {@link CSVFormat CSVFormat Javadoc}.
  * </p>
  *
  * <h2>Parsing into memory</h2>
@@ -126,24 +134,86 @@ import static org.apache.commons.csv.Token.Type.*;
  * Internal parser state is completely covered by the format and the reader-state.
  * </p>
  *
- * @version $Id: CSVParser.java 1742466 2016-05-05 19:55:37Z britter $
- *
  * @see <a href="package-summary.html">package documentation for more details</a>
  */
 public final class CSVParser implements Iterable<CSVRecord>, Closeable {
 
+    class CSVRecordIterator implements Iterator<CSVRecord> {
+        private CSVRecord current;
+
+        private CSVRecord getNextRecord() {
+            try {
+                return CSVParser.this.nextRecord();
+            } catch (final IOException e) {
+                throw new IllegalStateException(
+                        e.getClass().getSimpleName() + " reading next record: " + e.toString(), e);
+            }
+        }
+
+        @Override
+        public boolean hasNext() {
+            if (CSVParser.this.isClosed()) {
+                return false;
+            }
+            if (this.current == null) {
+                this.current = this.getNextRecord();
+            }
+
+            return this.current != null;
+        }
+
+        @Override
+        public CSVRecord next() {
+            if (CSVParser.this.isClosed()) {
+                throw new NoSuchElementException("CSVParser has been closed");
+            }
+            CSVRecord next = this.current;
+            this.current = null;
+
+            if (next == null) {
+                // hasNext() wasn't called before
+                next = this.getNextRecord();
+                if (next == null) {
+                    throw new NoSuchElementException("No more CSV records available");
+                }
+            }
+
+            return next;
+        }
+
+        @Override
+        public void remove() {
+            throw new UnsupportedOperationException();
+        }
+    }
+
+    /**
+     * Header information based on name and position.
+     */
+    private static final class Headers {
+        /**
+         * Header column positions (0-based)
+         */
+        final Map<String, Integer> headerMap;
+
+        /**
+         * Header names in column order
+         */
+        final List<String> headerNames;
+
+        Headers(final Map<String, Integer> headerMap, final List<String> headerNames) {
+            this.headerMap = headerMap;
+            this.headerNames = headerNames;
+        }
+    }
+
     /**
      * Creates a parser for the given {@link File}.
      *
-     * <p><strong>Note:</strong> This method internally creates a FileReader using
-     * {@link java.io.FileReader#FileReader(java.io.File)} which in turn relies on the default encoding of the JVM that
-     * is executing the code. If this is insufficient create a URL to the file and use
-     * {@link #parse(URL, Charset, CSVFormat)}</p>
-     *
      * @param file
      *            a CSV file. Must not be null.
      * @param charset
-     *            A charset
+     *            The Charset to decode the given file.
      * @param format
      *            the CSVFormat used for CSV parsing. Must not be null.
      * @return a new parser
@@ -153,11 +223,87 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
      *             If an I/O error occurs
      */
     public static CSVParser parse(final File file, final Charset charset, final CSVFormat format) throws IOException {
-        Assertions.notNull(file, "file");
-        Assertions.notNull(format, "format");
-        return new CSVParser(new InputStreamReader(new FileInputStream(file), charset), format);
+        Objects.requireNonNull(file, "file");
+        return parse(file.toPath(), charset, format);
     }
 
+    /**
+     * Creates a CSV parser using the given {@link CSVFormat}.
+     *
+     * <p>
+     * If you do not read all records from the given {@code reader}, you should call {@link #close()} on the parser,
+     * unless you close the {@code reader}.
+     * </p>
+     *
+     * @param inputStream
+     *            an InputStream containing CSV-formatted input. Must not be null.
+     * @param charset
+     *            The Charset to decode the given file.
+     * @param format
+     *            the CSVFormat used for CSV parsing. Must not be null.
+     * @return a new CSVParser configured with the given reader and format.
+     * @throws IllegalArgumentException
+     *             If the parameters of the format are inconsistent or if either reader or format are null.
+     * @throws IOException
+     *             If there is a problem reading the header or skipping the first record
+     * @since 1.5
+     */
+    @SuppressWarnings("resource")
+    public static CSVParser parse(final InputStream inputStream, final Charset charset, final CSVFormat format)
+            throws IOException {
+        Objects.requireNonNull(inputStream, "inputStream");
+        Objects.requireNonNull(format, "format");
+        return parse(new InputStreamReader(inputStream, charset), format);
+    }
+
+    /**
+     * Creates and returns a parser for the given {@link Path}, which the caller MUST close.
+     *
+     * @param path
+     *            a CSV file. Must not be null.
+     * @param charset
+     *            The Charset to decode the given file.
+     * @param format
+     *            the CSVFormat used for CSV parsing. Must not be null.
+     * @return a new parser
+     * @throws IllegalArgumentException
+     *             If the parameters of the format are inconsistent or if either file or format are null.
+     * @throws IOException
+     *             If an I/O error occurs
+     * @since 1.5
+     */
+    @SuppressWarnings("resource")
+    public static CSVParser parse(final Path path, final Charset charset, final CSVFormat format) throws IOException {
+        Objects.requireNonNull(path, "path");
+        Objects.requireNonNull(format, "format");
+        return parse(Files.newInputStream(path), charset, format);
+    }
+
+    /**
+     * Creates a CSV parser using the given {@link CSVFormat}
+     *
+     * <p>
+     * If you do not read all records from the given {@code reader}, you should call {@link #close()} on the parser,
+     * unless you close the {@code reader}.
+     * </p>
+     *
+     * @param reader
+     *            a Reader containing CSV-formatted input. Must not be null.
+     * @param format
+     *            the CSVFormat used for CSV parsing. Must not be null.
+     * @return a new CSVParser configured with the given reader and format.
+     * @throws IllegalArgumentException
+     *             If the parameters of the format are inconsistent or if either reader or format are null.
+     * @throws IOException
+     *             If there is a problem reading the header or skipping the first record
+     * @since 1.5
+     */
+    public static CSVParser parse(final Reader reader, final CSVFormat format) throws IOException {
+        return new CSVParser(reader, format);
+    }
+
+    // the following objects are shared to reduce garbage
+
     /**
      * Creates a parser for the given {@link String}.
      *
@@ -172,14 +318,14 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
      *             If an I/O error occurs
      */
     public static CSVParser parse(final String string, final CSVFormat format) throws IOException {
-        Assertions.notNull(string, "string");
-        Assertions.notNull(format, "format");
+        Objects.requireNonNull(string, "string");
+        Objects.requireNonNull(format, "format");
 
         return new CSVParser(new StringReader(string), format);
     }
 
     /**
-     * Creates a parser for the given URL.
+     * Creates and returns a parser for the given URL, which the caller MUST close.
      *
      * <p>
      * If you do not read all records from the given {@code url}, you should call {@link #close()} on the parser, unless
@@ -198,25 +344,25 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
      * @throws IOException
      *             If an I/O error occurs
      */
+    @SuppressWarnings("resource")
     public static CSVParser parse(final URL url, final Charset charset, final CSVFormat format) throws IOException {
-        Assertions.notNull(url, "url");
-        Assertions.notNull(charset, "charset");
-        Assertions.notNull(format, "format");
+        Objects.requireNonNull(url, "url");
+        Objects.requireNonNull(charset, "charset");
+        Objects.requireNonNull(format, "format");
 
         return new CSVParser(new InputStreamReader(url.openStream(), charset), format);
     }
 
-    // the following objects are shared to reduce garbage
-
     private final CSVFormat format;
 
-    /** A mapping of column names to column indices */
-    private final Map<String, Integer> headerMap;
+    private final Headers headers;
 
     private final Lexer lexer;
 
+    private final CSVRecordIterator csvRecordIterator;
+
     /** A record buffer for getRecord(). Grows as necessary and is reused. */
-    private final List<String> record = new ArrayList<String>();
+    private final List<String> recordList = new ArrayList<>();
 
     /**
      * The next record number to assign.
@@ -274,26 +420,27 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
      *             If there is a problem reading the header or skipping the first record
      * @since 1.1
      */
+    @SuppressWarnings("resource")
     public CSVParser(final Reader reader, final CSVFormat format, final long characterOffset, final long recordNumber)
-            throws IOException {
-        Assertions.notNull(reader, "reader");
-        Assertions.notNull(format, "format");
+        throws IOException {
+        Objects.requireNonNull(reader, "reader");
+        Objects.requireNonNull(format, "format");
 
-        this.format = format;
+        this.format = format.copy();
         this.lexer = new Lexer(format, new ExtendedBufferedReader(reader));
-        this.headerMap = this.initializeHeader();
+        this.csvRecordIterator = new CSVRecordIterator();
+        this.headers = createHeaders();
         this.characterOffset = characterOffset;
         this.recordNumber = recordNumber - 1;
     }
 
-    private void addRecordValue(boolean lastRecord) {
+    private void addRecordValue(final boolean lastRecord) {
         final String input = this.reusableToken.content.toString();
         final String inputClean = this.format.getTrim() ? input.trim() : input;
         if (lastRecord && inputClean.isEmpty() && this.format.getTrailingDelimiter()) {
             return;
         }
-        final String nullString = this.format.getNullString();
-        this.record.add(inputClean.equals(nullString) ? null : inputClean);
+        this.recordList.add(handleNull(inputClean));
     }
 
     /**
@@ -309,6 +456,73 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
         }
     }
 
+    private Map<String, Integer> createEmptyHeaderMap() {
+        return this.format.getIgnoreHeaderCase() ?
+                new TreeMap<>(String.CASE_INSENSITIVE_ORDER) :
+                new LinkedHashMap<>();
+    }
+
+    /**
+     * Creates the name to index mapping if the format defines a header.
+     *
+     * @return null if the format has no header.
+     * @throws IOException if there is a problem reading the header or skipping the first record
+     */
+    private Headers createHeaders() throws IOException {
+        Map<String, Integer> hdrMap = null;
+        List<String> headerNames = null;
+        final String[] formatHeader = this.format.getHeader();
+        if (formatHeader != null) {
+            hdrMap = createEmptyHeaderMap();
+            String[] headerRecord = null;
+            if (formatHeader.length == 0) {
+                // read the header from the first line of the file
+                final CSVRecord nextRecord = this.nextRecord();
+                if (nextRecord != null) {
+                    headerRecord = nextRecord.values();
+                }
+            } else {
+                if (this.format.getSkipHeaderRecord()) {
+                    this.nextRecord();
+                }
+                headerRecord = formatHeader;
+            }
+
+            // build the name to index mappings
+            if (headerRecord != null) {
+                for (int i = 0; i < headerRecord.length; i++) {
+                    final String header = headerRecord[i];
+                    final boolean emptyHeader = header == null || header.trim().isEmpty();
+                    if (emptyHeader && !this.format.getAllowMissingColumnNames()) {
+                        throw new IllegalArgumentException(
+                            "A header name is missing in " + Arrays.toString(headerRecord));
+                    }
+                    // Note: This will always allow a duplicate header if the header is empty
+                    final boolean containsHeader = header != null && hdrMap.containsKey(header);
+                    if (containsHeader && !emptyHeader && !this.format.getAllowDuplicateHeaderNames()) {
+                        throw new IllegalArgumentException(
+                            String.format(
+                                "The header contains a duplicate name: \"%s\" in %s. If this is valid then use CSVFormat.withAllowDuplicateHeaderNames().",
+                                header, Arrays.toString(headerRecord)));
+                    }
+                    if (header != null) {
+                        hdrMap.put(header, Integer.valueOf(i));
+                        if (headerNames == null) {
+                            headerNames = new ArrayList<>(headerRecord.length);
+                        }
+                        headerNames.add(header);
+                    }
+                }
+            }
+        }
+        if (headerNames == null) {
+            headerNames = Collections.emptyList(); //immutable
+        } else {
+            headerNames = Collections.unmodifiableList(headerNames);
+        }
+        return new Headers(hdrMap, headerNames);
+    }
+
     /**
      * Returns the current line number in the input stream.
      *
@@ -324,14 +538,59 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
     }
 
     /**
-     * Returns a copy of the header map that iterates in column order.
+     * Gets the first end-of-line string encountered.
+     *
+     * @return the first end-of-line string
+     * @since 1.5
+     */
+    public String getFirstEndOfLine() {
+        return lexer.getFirstEol();
+    }
+
+    /**
+     * Returns a copy of the header map.
      * <p>
      * The map keys are column names. The map values are 0-based indices.
      * </p>
-     * @return a copy of the header map that iterates in column order.
+     * <p>
+     * Note: The map can only provide a one-to-one mapping when the format did not
+     * contain null or duplicate column names.
+     * </p>
+     *
+     * @return a copy of the header map.
      */
     public Map<String, Integer> getHeaderMap() {
-        return this.headerMap == null ? null : new LinkedHashMap<String, Integer>(this.headerMap);
+        if (this.headers.headerMap == null) {
+            return null;
+        }
+        final Map<String, Integer> map = createEmptyHeaderMap();
+        map.putAll(this.headers.headerMap);
+        return map;
+    }
+
+    /**
+     * Returns the header map.
+     *
+     * @return the header map.
+     */
+    Map<String, Integer> getHeaderMapRaw() {
+        return this.headers.headerMap;
+    }
+
+    /**
+     * Returns a read-only list of header names that iterates in column order.
+     * <p>
+     * Note: The list provides strings that can be used as keys in the header map.
+     * The list will not contain null column names if they were present in the input
+     * format.
+     * </p>
+     *
+     * @return read-only list of header names that iterates in column order.
+     * @see #getHeaderMap()
+     * @since 1.7
+     */
+    public List<String> getHeaderNames() {
+        return Collections.unmodifiableList(headers.headerNames);
     }
 
     /**
@@ -362,7 +621,7 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
      */
     public List<CSVRecord> getRecords() throws IOException {
         CSVRecord rec;
-        final List<CSVRecord> records = new ArrayList<CSVRecord>();
+        final List<CSVRecord> records = new ArrayList<>();
         while ((rec = this.nextRecord()) != null) {
             records.add(rec);
         }
@@ -370,52 +629,26 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
     }
 
     /**
-     * Initializes the name to index mapping if the format defines a header.
+     * Handle whether input is parsed as null
      *
-     * @return null if the format has no header.
-     * @throws IOException if there is a problem reading the header or skipping the first record
+     * @param input
+     *           the cell data to further processed
+     * @return null if input is parsed as null, or input itself if input isn't parsed as null
      */
-    private Map<String, Integer> initializeHeader() throws IOException {
-        Map<String, Integer> hdrMap = null;
-        final String[] formatHeader = this.format.getHeader();
-        if (formatHeader != null) {
-            hdrMap = this.format.getIgnoreHeaderCase() ?
-                    new TreeMap<String, Integer>(String.CASE_INSENSITIVE_ORDER) :
-                    new LinkedHashMap<String, Integer>();
-
-            String[] headerRecord = null;
-            if (formatHeader.length == 0) {
-                // read the header from the first line of the file
-                final CSVRecord nextRecord = this.nextRecord();
-                if (nextRecord != null) {
-                    headerRecord = nextRecord.values();
-                }
-            } else {
-                if (this.format.getSkipHeaderRecord()) {
-                    this.nextRecord();
-                }
-                headerRecord = formatHeader;
-            }
-
-            // build the name to index mappings
-            if (headerRecord != null) {
-                for (int i = 0; i < headerRecord.length; i++) {
-                    final String header = headerRecord[i];
-                    final boolean containsHeader = hdrMap.containsKey(header);
-                    final boolean emptyHeader = header == null || header.trim().isEmpty();
-                    if (containsHeader && (!emptyHeader || !this.format.getAllowMissingColumnNames())) {
-                        throw new IllegalArgumentException("The header contains a duplicate name: \"" + header +
-                                "\" in " + Arrays.toString(headerRecord));
-                    }
-                    hdrMap.put(header, Integer.valueOf(i));
-                }
-            }
+    private String handleNull(final String input) {
+        final boolean isQuoted = this.reusableToken.isQuoted;
+        final String nullString = format.getNullString();
+        final boolean strictQuoteMode = isStrictQuoteMode();
+        if (input.equals(nullString)) {
+            // nullString = NULL(String), distinguish between "NULL" and NULL in ALL_NON_NULL or NON_NUMERIC quote mode
+            return strictQuoteMode && isQuoted ? input : null;
         }
-        return hdrMap;
+        // don't set nullString, distinguish between "" and ,, (absent values) in All_NON_NULL or NON_NUMERIC quote mode
+        return strictQuoteMode && nullString == null && input.isEmpty() && !isQuoted ? null : input;
     }
 
     /**
-     * Gets whether this parser is closed.
+     * Tests whether this parser is closed.
      *
      * @return whether this parser is closed.
      */
@@ -424,63 +657,31 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
     }
 
     /**
-     * Returns an iterator on the records.
+     * Tests whether the format's {@link QuoteMode} is {@link QuoteMode#ALL_NON_NULL} or {@link QuoteMode#NON_NUMERIC}.
      *
-     * <p>IOExceptions occurring during the iteration are wrapped in a
-     * RuntimeException.
-     * If the parser is closed a call to {@code next()} will throw a
-     * NoSuchElementException.</p>
+     * @return true if the format's {@link QuoteMode} is {@link QuoteMode#ALL_NON_NULL} or
+     *         {@link QuoteMode#NON_NUMERIC}.
+     */
+    private boolean isStrictQuoteMode() {
+        return this.format.getQuoteMode() == QuoteMode.ALL_NON_NULL ||
+               this.format.getQuoteMode() == QuoteMode.NON_NUMERIC;
+    }
+
+    /**
+     * Returns the record iterator.
+     *
+     * <p>
+     * An {@link IOException} caught during the iteration are re-thrown as an
+     * {@link IllegalStateException}.
+     * </p>
+     * <p>
+     * If the parser is closed a call to {@link Iterator#next()} will throw a
+     * {@link NoSuchElementException}.
+     * </p>
      */
     @Override
     public Iterator<CSVRecord> iterator() {
-        return new Iterator<CSVRecord>() {
-            private CSVRecord current;
-
-            private CSVRecord getNextRecord() {
-                try {
-                    return CSVParser.this.nextRecord();
-                } catch (final IOException e) {
-                    // TODO: This is not great, throw an ISE instead?
-                    throw new RuntimeException(e);
-                }
-            }
-
-            @Override
-            public boolean hasNext() {
-                if (CSVParser.this.isClosed()) {
-                    return false;
-                }
-                if (this.current == null) {
-                    this.current = this.getNextRecord();
-                }
-
-                return this.current != null;
-            }
-
-            @Override
-            public CSVRecord next() {
-                if (CSVParser.this.isClosed()) {
-                    throw new NoSuchElementException("CSVParser has been closed");
-                }
-                CSVRecord next = this.current;
-                this.current = null;
-
-                if (next == null) {
-                    // hasNext() wasn't called before
-                    next = this.getNextRecord();
-                    if (next == null) {
-                        throw new NoSuchElementException("No more CSV records available");
-                    }
-                }
-
-                return next;
-            }
-
-            @Override
-            public void remove() {
-                throw new UnsupportedOperationException();
-            }
-        };
+        return csvRecordIterator;
     }
 
     /**
@@ -492,7 +693,7 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
      */
     CSVRecord nextRecord() throws IOException {
         CSVRecord result = null;
-        this.record.clear();
+        this.recordList.clear();
         StringBuilder sb = null;
         final long startCharPosition = lexer.getCharacterPosition() + this.characterOffset;
         do {
@@ -526,13 +727,23 @@ public final class CSVParser implements Iterable<CSVRecord>, Closeable {
             }
         } while (this.reusableToken.type == TOKEN);
 
-        if (!this.record.isEmpty()) {
+        if (!this.recordList.isEmpty()) {
             this.recordNumber++;
             final String comment = sb == null ? null : sb.toString();
-            result = new CSVRecord(this.record.toArray(new String[this.record.size()]), this.headerMap, comment,
-                    this.recordNumber, startCharPosition);
+            result = new CSVRecord(this, this.recordList.toArray(Constants.EMPTY_STRING_ARRAY), comment,
+                this.recordNumber, startCharPosition);
         }
         return result;
     }
 
+    /**
+     * Returns a sequential {@code Stream} with this collection as its source.
+     *
+     * @return a sequential {@code Stream} with this collection as its source.
+     * @since 1.9.0
+     */
+    public Stream<CSVRecord> stream() {
+        return StreamSupport.stream(Spliterators.spliteratorUnknownSize(iterator(), Spliterator.ORDERED), false);
+    }
+
 }
